@@ -10,14 +10,18 @@ import numpy
 from nion.data import Calibration
 from nion.data import DataAndMetadata
 from nion.swift import Application
-from nion.swift import DocumentController
 from nion.swift import Facade
 from nion.swift.model import DataItem
-from nion.swift.model import DocumentModel
+from nion.swift.model import Symbolic
+from nion.swift.test import TestContext
 from nion.ui import TestUI
 
+from nion.eels_analysis import eels_analysis
 from nion.eels_analysis import PeriodicTable
+
 from nionswift_plugin.nion_eels_analysis import ElementalMappingController
+from nionswift_plugin.nion_eels_analysis import AlignZLP
+from nionswift_plugin.nion_eels_analysis import ThicknessMap
 
 
 Facade.initialize()
@@ -32,7 +36,10 @@ class TestElementalMappingController(unittest.TestCase):
         pass
 
     def __create_spectrum_image(self) -> DataItem.DataItem:
-        data = numpy.zeros((8, 8, 1024))
+        return DataItem.new_data_item(self.__create_spectrum_image_xdata())
+
+    def __create_spectrum_image_xdata(self, dtype=numpy.float32):
+        data = numpy.zeros((8, 8, 1024), dtype)
         for row in range(data.shape[0]):
             for column in range(data.shape[1]):
                 data[row, column, :] = numpy.random.uniform(10, 1000, 1024)
@@ -40,10 +47,10 @@ class TestElementalMappingController(unittest.TestCase):
         dimensional_calibrations = [Calibration.Calibration(units="nm"), Calibration.Calibration(units="nm"), Calibration.Calibration(scale=2.0, units="eV")]
         data_descriptor = DataAndMetadata.DataDescriptor(is_sequence=False, collection_dimension_count=2, datum_dimension_count=1)
         xdata = DataAndMetadata.new_data_and_metadata(data, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, data_descriptor=data_descriptor)
-        return DataItem.new_data_item(xdata)
+        return xdata
 
     def __create_spectrum(self) -> DataItem.DataItem:
-        data = numpy.random.uniform(10, 1000, 1024)
+        data = numpy.random.uniform(10, 1000, 1024).astype(numpy.float32)
         intensity_calibration = Calibration.Calibration(units="~")
         dimensional_calibrations = [Calibration.Calibration(scale=2.0, units="eV")]
         data_descriptor = DataAndMetadata.DataDescriptor(is_sequence=False, collection_dimension_count=0, datum_dimension_count=1)
@@ -76,10 +83,10 @@ class TestElementalMappingController(unittest.TestCase):
             time.sleep(1/50)
 
     def test_explore_creates_initial_line_plot(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             data_item = self.__create_spectrum_image()
             document_model.append_data_item(data_item)
             elemental_mapping_controller.set_current_data_item(data_item)
@@ -94,10 +101,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual("explore", explorer_display_item.graphics[-1].graphic_id)
 
     def test_explore_adds_edge(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -111,7 +118,12 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertIsNotNone(elemental_mapping_controller.model_data_item)
             energy_calibration = explorer_data_item.dimensional_calibrations[-1]
             explorer_display_item.graphics[-1].interval = energy_calibration.convert_from_calibrated_value(1200) / 1024, energy_calibration.convert_from_calibrated_value(1226) / 1024
-            document_controller.periodic()  # update explorer interval
+            for _ in range(3):
+                # there is something funny about how async works; recent versions of Swift are faster
+                # and have revealed some race condition about how items get added to the async queue.
+                # to avoid that problem, do periodic over a period of a few ms.
+                document_controller.periodic()  # update explorer interval
+                time.sleep(0.01)
             edges = PeriodicTable.PeriodicTable().find_edges_in_energy_interval(elemental_mapping_controller.explorer_interval)
             elemental_mapping_controller.add_edge(edges[0])
             self.assertEqual(1, len(document_model.data_structures))
@@ -123,10 +135,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(3, edge_data_struct.get_property_value("subshell_index"))
 
     def test_adding_multiple_edges(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -135,10 +147,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(2, len(document_model.data_structures))
 
     def test_removing_edges(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -152,10 +164,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(0, len(document_model.data_structures))
 
     def test_controller_has_proper_edge_bundles_when_explorer_selected(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -169,10 +181,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(1, len(edge_bundle))
 
     def test_picking_edge_produces_properly_configured_composite(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             model_display_item = document_model.get_display_item_for_data_item(model_data_item)
@@ -189,10 +201,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(document_model.data_structures[0], document_model.data_structures[1].get_referenced_object("edge"))
 
     def test_deleting_pick_also_deletes_computation(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -210,10 +222,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(1, len(document_model.data_structures))
 
     def test_deleting_pick_region_also_deletes_pick_composition(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             model_display_item = document_model.get_display_item_for_data_item(model_data_item)
@@ -233,10 +245,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(1, len(document_model.data_structures))
 
     def test_selecting_composite_updates_edge_value(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -251,11 +263,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(si_edge.data_structure, elemental_mapping_controller.edge.data_structure)
 
     def test_background_subtraction_computation_functions_reasonably(self):
-        document_model = DocumentModel.DocumentModel()
-        self.app._set_document_model(document_model)  # required to allow API to find document model
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller_with_application()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -265,9 +276,9 @@ class TestElementalMappingController(unittest.TestCase):
             spectrum_display_item = document_model.get_display_item_for_data_item(spectrum_data_item)
             spectrum_display_data_channel = spectrum_display_item.get_display_data_channel_for_data_item(spectrum_data_item)
             computation = document_model.create_computation()
-            computation.create_object("eels_spectrum_xdata", document_model.get_object_specifier(spectrum_display_data_channel, "display_xdata"))
-            computation.create_input("fit_interval", document_model.get_object_specifier(si_edge.data_structure), "fit_interval")
-            computation.create_input("signal_interval", document_model.get_object_specifier(si_edge.data_structure), "signal_interval")
+            computation.create_input_item("eels_spectrum_xdata", Symbolic.make_item(spectrum_display_data_channel, type="display_xdata"))
+            computation.create_input_item("fit_interval", Symbolic.make_item(si_edge.data_structure), property_name="fit_interval")
+            computation.create_input_item("signal_interval", Symbolic.make_item(si_edge.data_structure), property_name="signal_interval")
             computation.processing_id = "eels.background_subtraction"
             document_model.append_computation(computation)
             document_model.recompute_all()
@@ -275,10 +286,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(2, len(document_model.data_items))
 
     def test_changing_edge_configures_other_items_correctly(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             model_display_item = document_model.get_display_item_for_data_item(model_data_item)
@@ -313,23 +324,23 @@ class TestElementalMappingController(unittest.TestCase):
             old_edge_data_structure.set_property_value("signal_interval", new_signal_interval)
             self.assertEqual(old_fit_interval, eels_display_item.graphics[0].interval)
             self.assertEqual(old_signal_interval, eels_display_item.graphics[1].interval)
-            self.assertEqual(old_fit_interval, computation._get_variable("fit_interval").bound_item.value)
-            self.assertEqual(old_signal_interval, computation._get_variable("signal_interval").bound_item.value)
+            self.assertEqual(old_fit_interval, computation.get_input("fit_interval"))
+            self.assertEqual(old_signal_interval, computation.get_input("signal_interval"))
             # ensure changing new edge affects all connections
             new_edge_data_structure.set_property_value("fit_interval", new_fit_interval)
             new_edge_data_structure.set_property_value("signal_interval", new_signal_interval)
             self.assertEqual(new_fit_interval, eels_display_item.graphics[0].interval)
             self.assertEqual(new_signal_interval, eels_display_item.graphics[1].interval)
-            self.assertEqual(new_fit_interval, computation._get_variable("fit_interval").bound_item.value)
-            self.assertEqual(new_signal_interval, computation._get_variable("signal_interval").bound_item.value)
+            self.assertEqual(new_fit_interval, computation.get_input("fit_interval"))
+            self.assertEqual(new_signal_interval, computation.get_input("signal_interval"))
             # and the edge reference
             self.assertEqual(new_edge_data_structure, edge_ref_data_structure.get_referenced_object("edge"))
 
     def test_mapping_edge_produces_properly_configured_map(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -345,10 +356,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertEqual(mapped_data_item.dimensional_calibrations, model_data_item.dimensional_calibrations[0:2])
 
     def test_multiprofile_of_two_maps_builds_two_line_profiles(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -370,10 +381,10 @@ class TestElementalMappingController(unittest.TestCase):
             self.assertIn(line_profile2_data_item, composite_display_item.data_items)
 
     def test_multiprofile_of_two_maps_connects_line_profiles(self):
-        document_model = DocumentModel.DocumentModel()
-        elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
-        document_controller = DocumentController.DocumentController(self.app.ui, document_model, workspace_id="library")
-        with contextlib.closing(document_controller), contextlib.closing(elemental_mapping_controller):
+        with TestContext.create_memory_context() as test_context:
+            document_controller = test_context.create_document_controller()
+            document_model = document_controller.document_model
+            elemental_mapping_controller = ElementalMappingController.ElementalMappingController(document_model)
             model_data_item = self.__create_spectrum_image()
             document_model.append_data_item(model_data_item)
             elemental_mapping_controller.set_current_data_item(model_data_item)
@@ -397,3 +408,27 @@ class TestElementalMappingController(unittest.TestCase):
             line_region1.vector = (0.11, 0.12), (0.21, 0.22)
             self.assertEqual(line_region1.vector, line_region2.vector)
             self.assertEqual(line_region1.width, line_region2.width)
+
+    def test_map_background_subtracted_signal_keeps_input_dtype(self):
+        si_xdata_32 = self.__create_spectrum_image_xdata(dtype=numpy.float32)
+        si_xdata_64 = self.__create_spectrum_image_xdata(dtype=numpy.float64)
+        mapped_xdata_32 = eels_analysis.map_background_subtracted_signal(si_xdata_32, None, [(0.2, 0.3)], (0.4, 0.5))
+        mapped_xdata_64 = eels_analysis.map_background_subtracted_signal(si_xdata_64, None, [(0.2, 0.3)], (0.4, 0.5))
+        self.assertEqual(numpy.float32, mapped_xdata_32.data.dtype)
+        self.assertEqual(numpy.float64, mapped_xdata_64.data.dtype)
+
+    def test_align_zlp_keeps_input_dtype(self):
+        si_xdata_32 = self.__create_spectrum_image_xdata(dtype=numpy.float32)
+        si_xdata_64 = self.__create_spectrum_image_xdata(dtype=numpy.float64)
+        mapped_xdata_32 = AlignZLP.align_zlp_xdata(si_xdata_32)[0]
+        mapped_xdata_64 = AlignZLP.align_zlp_xdata(si_xdata_64)[0]
+        self.assertEqual(numpy.float32, mapped_xdata_32.data.dtype)
+        self.assertEqual(numpy.float64, mapped_xdata_64.data.dtype)
+
+    def test_map_thickness_is_always_float32(self):
+        si_xdata_32 = self.__create_spectrum_image_xdata(dtype=numpy.float32)
+        si_xdata_64 = self.__create_spectrum_image_xdata(dtype=numpy.float64)
+        mapped_xdata_32 = ThicknessMap.map_thickness_xdata(si_xdata_32)
+        mapped_xdata_64 = ThicknessMap.map_thickness_xdata(si_xdata_64)
+        self.assertEqual(numpy.float32, mapped_xdata_32.data.dtype)
+        self.assertEqual(numpy.float32, mapped_xdata_64.data.dtype)
